"""
Exceptions raised by cli-toolkit
"""


class ConfigurationError(Exception):
    """
    Errors raised by configuration processing
    """


class ScriptError(Exception):
    """
    Errors raise during script processing
    """
