"""
CLI utility toolkit

This module implements python modules to implement command line tools,
extending python argparse.ArgumentParser to make writing nested CLI
commands easier.
"""
__version__ = '1.2.0'
